from datetime import date, datetime, time
from decimal import Decimal


__all__ = [
    "escape_any", "escape_string", "escape_datetime", "escape_date",
    "escape_time", "escape_boolean", "escape_float",
]


def escape_any(value):
    """
    Section 4.1.2 defines SPARQL shortened forms
    https://www.w3.org/TR/2013/REC-sparql11-query-20130321/#QSynLiterals

    Examples of literal syntax in SPARQL include:

        "chat"
        'chat'@fr with language tag "fr"
        "xyz"^^<http://example.org/ns/userDatatype>
        "abc"^^appNS:appDataType
        '''The librarian said, "Perhaps you would enjoy 'War and Peace'."'''
        1, which is the same as "1"^^xsd:integer
        1.3, which is the same as "1.3"^^xsd:decimal
        1.300, which is the same as "1.300"^^xsd:decimal
        1.0e6, which is the same as "1.0e6"^^xsd:double
        true, which is the same as "true"^^xsd:boolean
        false, which is the same as "false"^^xsd:boolean
    """
    if isinstance(value, type):
        raise TypeError("object %r is not an instance" % value)
    for type_, escape_method in escapers:
        if isinstance(value, type_):
            return escape_method(value)
    return escape_string(str(value))


_string_replacements = [
    # always replace first the \ to avoid doubling the future ones
    ('\\', '\\\\'),
    ('"', '\\"'),
    ('\n', "\\n"),
    ('\r', "\\r"),
]


def escape_string(value):
    for old, new in _string_replacements:
        value = value.replace(old, new)
    return '"%s"' % value


def escape_datetime(value):
    return '"%s"^^xsd:dateTime' % value.isoformat()


def escape_date(value):
    return '"%s"^^xsd:date' % value.isoformat()


def escape_time(value):
    return '"%s"^^xsd:time' % value.isoformat()


def escape_boolean(value):
    return "true" if value else "false"


def escape_float(value):
    """
    Python's float are usually double:
    https://docs.python.org/3.6/library/stdtypes.html#numeric-types-int-float-complex
    """
    return '"%s"^^xsd:double' % value


escapers = [
    (bool, escape_boolean),
    (datetime, escape_datetime),
    (date, escape_date),
    (time, escape_time),
    (float, escape_float),
    (int, str),
    (Decimal, str),
]



from itertools import groupby
import re
from textwrap import indent

from .escape import escape_any, escape_string, escapers


__all__ = [
    "RDFTerm", "Node", "Triples", "PrefixedName", "IRI", "Literal", "UNDEF",
    "Namespace", "RDF",
]


class RDFTerm:
    """
    A base class for any class that can be print (using str()) and return a
    valid RDF term (i.e. that can be used in a SPARQL query).
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        if isinstance(other, RDFTerm):
            return self.value == other.value
        else:
            return self.value == other

    def __hash__(self):
        return hash(self.value)


class Node(list):
    """
    A magical list of tuples (p, o) with a subject that can be printed to a
    SPARQL query.

    It can also accept a dict in argument.
    """
    def __init__(self, subject, value=[]):
        self.subject = subject
        if isinstance(value, dict):
            super(Node, self).__init__(value.items())
        else:
            super(Node, self).__init__(value)

    def __str__(self):
        return "".join(self._output_triples())

    def _output_triples(self):
        extra_nodes = []
        it = iter(self)
        p, o = next(it)
        yield "%s %s %s" % (self.subject, p, escape_any(o))
        if isinstance(o, Node):
            extra_nodes.append(o)
        for p, o in it:
            assert p is not None, "predicate not defined"
            yield " ;\n"
            yield "    %s %s" % (p, escape_any(o))
            if isinstance(o, Node):
                extra_nodes.append(o)
        yield " ."
        for node in sorted(extra_nodes, key=lambda x: str(x.subject)):
            yield "\n\n"
            yield str(node)

    def _group_key(self, x):
        return (str(x[0]), str(x[1]))

    def __iter__(self):
        for p, o in sorted(super().__iter__(), key=self._group_key):
            if o is None:
                continue
            yield (p, o)


class Triples(list):
    """
    A magical list of tuples (s, p, o) that be printed to a SPARQL query.
    """
    def __init__(self, value=[]):
        assert all(isinstance(x, (tuple, Node)) for x in value), \
            "only tuples and Node are accepted, received: %r" % value
        assert all(len(x) == 3 for x in value if isinstance(x, tuple)), \
            "tuples must be of length 3"
        return super(Triples, self).__init__(value)

    def append(self, triple):
        assert isinstance(triple, (tuple, Node)), \
            "only tuples are Node accepted, received: %r" % triple
        assert len(triple) == 3 if isinstance(triple, tuple) else True, \
            "tuple must be of length 3"
        return super(Triples, self).append(triple)

    def extend(self, value):
        assert all(isinstance(x, (tuple, Node)) for x in value), \
            "only tuples are Node accepted, received: %r" % value
        assert all(len(x) == 3 for x in value if isinstance(x, tuple)), \
            "tuples must be of length 3"
        return super(Triples, self).extend(value)

    def __str__(self):
        return "".join(self._output_triples())

    def indent(self, spaces):
        return indent(str(self), spaces)

    def _output_triples(self):
        item = None
        for s, group in groupby(self, self._group_key):
            assert s is not None, "subject not defined"
            if item is None:
                pass
            elif isinstance(item, tuple):
                yield " .\n\n"
            elif isinstance(item, Node):
                yield "\n\n"
            item = next(group)
            if isinstance(item, tuple):
                s, p, o = item
                yield "%s %s %s" % (s, p, escape_any(o))
                for _, p, o in group:
                    assert p is not None, "predicate not defined"
                    if o is None:
                        continue
                    yield " ;\n"
                    yield "    %s %s" % (p, escape_any(o))
            elif isinstance(item, Node):
                yield str(item)
        if isinstance(item, tuple):
            yield " ."

    def _group_key(self, x):
        if isinstance(x, tuple):
            return x[0]
        elif isinstance(x, Node):
            return id(x)


class PrefixedName(RDFTerm):
    def __init__(self, base_iri, prefix_label, local_part):
        self.base_iri = base_iri
        self.prefix_label = prefix_label
        self.local_part = local_part

    def __str__(self):
        return "%s:%s" % (self.prefix_label, self.local_part)

    def __repr__(self):
        return "<PrefixedName %s>" % self

    def __eq__(self, other):
        if isinstance(other, PrefixedName):
            return (self.prefix_label == other.prefix_label and
                    self.local_part == other.local_part)
        else:
            return self.iri() == other

    def __hash__(self):
        return hash(self.iri())

    def iri(self):
        return IRI(self.base_iri.value + self.local_part)


class IRI(RDFTerm):
    __re_invalid_chars__ = re.compile('[<>"{}|^`[-\\]\x00-\x20]')

    def __init__(self, value):
        self.value = value
        self.ref = "<%s>" % self.__re_invalid_chars__.sub(
            lambda x: '%{:02X}'.format(ord(x.group(0))), value)

    def __str__(self):
        return self.ref

    def __repr__(self):
        return "<IRI %s>" % self.value

    def __eq__(self, other):
        if isinstance(other, IRI):
            return self.value == other.value
        else:
            return self.value == other

    def __hash__(self):
        return hash(self.value)

    def __add__(self, other):
        return IRI("%s%s" % (self.value, other))


class Literal(RDFTerm):
    def __init__(self, value, lang=None):
        self.value = value
        self.lang = lang

    def __str__(self):
        return "%s@%s" % (escape_string(self.value), self.lang)

    def __eq__(self, other):
        if isinstance(other, Literal):
            return (self.value == other.value and self.lang == other.lang)
        else:
            return self.value == other

    def __hash__(self):
        return hash((self.value, self.lang))


class UNDEF(RDFTerm):
    def __init__(self):
        pass

    def __str__(self):
        return "UNDEF"

    def __eq__(self, other):
        return isinstance(other, UNDEF)

    def __hash__(self):
        return hash(UNDEF)


class MetaNamespace(type):
    prefixes = {}

    def __new__(mcs, name, bases, nmspc):
        if bases:
            assert '__iri__' in nmspc, \
                "missing attribute __iri__ for class %s" % name
            iri = nmspc['__iri__']
            prefix_label = nmspc.get('__prefix_label__', name.lower())
            nmspc = {
                k: (
                    PrefixedName(iri, prefix_label, k)
                    if v is PrefixedName else v
                )
                for k, v in nmspc.items()
            }
            nmspc['__prefix_label__'] = prefix_label
        return super(MetaNamespace, mcs).__new__(mcs, name, bases, nmspc)

    def __init__(cls, name, bases, nmspc):
        super(MetaNamespace, cls).__init__(name, bases, nmspc)
        if bases:
            MetaNamespace.prefixes[name] = cls


class Namespace(metaclass=MetaNamespace):
    __prefix_label__ = ""


class RDF(Namespace):
    __iri__ = IRI("http://www.w3.org/1999/02/22-rdf-syntax-ns#")

    HTML = PrefixedName
    langString = PrefixedName
    PlainLiteral = PrefixedName
    type = PrefixedName
    Property = PrefixedName
    Statement = PrefixedName
    subject = PrefixedName
    predicate = PrefixedName
    object = PrefixedName
    Bag = PrefixedName
    Seq = PrefixedName
    Alt = PrefixedName
    value = PrefixedName
    List = PrefixedName
    nil = PrefixedName
    first = PrefixedName
    rest = PrefixedName
    XMLLiteral = PrefixedName


escapers.extend([
    (Node, lambda x: x.subject),
    (RDFTerm, str),
])



import aiohttp
import logging
from math import ceil, log10
from os import environ as ENV
import re
from string import Formatter
from textwrap import dedent, indent
from typing import List, Optional

from .syntax import IRI, MetaNamespace, Namespace

__all__ = ['SPARQLClient', 'SPARQLRequestFailed', 'SPARQLQueryFormatter']


logger = logging.getLogger(__name__)


class SPARQLRequestFailed(aiohttp.client_exceptions.ClientResponseError):
    def __init__(self, request_info, history, *,
                 code=0, message='', headers=None, explanation=None):
        super(SPARQLRequestFailed, self).__init__(
            request_info, history, code=code, message=message, headers=headers)
        self.explanation = explanation

    def __str__(self):
        base_message = super(SPARQLRequestFailed, self).__str__()
        return "%s, explanation=%r" % (base_message, self.explanation)


class SPARQLQueryFormatter(Formatter):
    """
    This custom formatter redefine the parse method of the default Python's
    formatter in order to use {{}} instead of {} as replacement field.

    It also automatically indents text, example:
    s = "\n".join(["b", "c", "d"])
    SPARQLQueryFormatter().vformat("a\n  {{}}\ne", [s], {})
    # returns:
    # a
    #     b
    #     c
    #     d
    # e
    """
    re_token = \
        re.compile(r"((?:[^{]+|\{[^{]+|\{$)*)(?:\{\{((?:[^}]+|}[^}])*)\}\})?")
    re_field = re.compile(r"(.*)(?:!([sra]))?(?::([^{}]*))?")
    re_indent = re.compile(r"(.*)^(\s*)$", flags=(re.M + re.S))
    indent = ""

    def parse(self, s):
        if not s:
            return
        else:
            for match in self.re_token.finditer(s):
                imatch = self.re_indent.fullmatch(match.group(1))
                if imatch:
                    text = imatch.group(1)
                    self.indent = imatch.group(2)
                else:
                    text = match.group(1)
                    self.indent = ""
                if match.group(2) is None:
                    if match.end() != len(s):
                        raise Exception(
                            "Not terminated token: %r" % s[match.end():])
                    yield (text, None, None, None)
                    break
                else:
                    fmatch = self.re_field.fullmatch(match.group(2))
                    if not fmatch:
                        raise Exception(
                            "Cannot parse token: %r" % match.group(2))
                    yield (text, fmatch.group(1), fmatch.group(3),
                           fmatch.group(2))

    def format_field(self, value, format_spec):
        return indent(format(value, format_spec), self.indent)


class SPARQLClient(aiohttp.ClientSession):
    def __init__(self,
                 endpoint: Optional[str] = None,
                 update_endpoint: Optional[str] = None,
                 prefixes: Optional[List[Namespace]] = None,
                 graph: Optional[str] = None,
                 **kwargs):
        super(SPARQLClient, self).__init__(**kwargs)
        self.endpoint = endpoint or ENV['MU_SPARQL_ENDPOINT']
        self.graph = graph or IRI(ENV['MU_APPLICATION_GRAPH'])
        self.update_endpoint = update_endpoint or self.endpoint
        self._generate_prefixes(prefixes)

    def _generate_prefixes(self, prefixes):
        header = [
            "PREFIX %s: %s" % (x.__prefix_label__, x.__iri__)
            for x in sorted(MetaNamespace.prefixes.values(),
                            key=lambda x: x.__prefix_label__)
        ]
        if prefixes:
            header.append("")
            header.extend([
                "PREFIX %s: %s" % (x[0], x[1])
                for x in sorted(prefixes.items(), key=lambda x: x[0])
            ])
        self._prefixes_header = "\n".join(header) + "\n"

    def _prepare_query(self, query: str, *args, **keywords) -> dict:
        lines = [self._prefixes_header]
        lines.extend([dedent(query).strip()])
        query_args = {
            'graph': self.graph,
        }
        query_args.update(keywords)
        query_formatter = SPARQLQueryFormatter()
        return query_formatter.vformat("\n".join(lines), args, query_args)

    def _pretty_print_query(self, query: str) -> str:
        query = query.rstrip()
        ln_indent = ceil(log10(query.count("\n")))
        return "\n".join(
            ("{:-%dd}: {}" % ln_indent).format(i, x)
            for i, x in enumerate(query.split("\n"), 1)
        )

    async def query(self, query: str, *args, **keywords) -> dict:
        headers = {
            "Accept": "application/json",
        }
        full_query = self._prepare_query(query, *args, **keywords)
        logger.debug("Sending SPARQL query to %s: \n%s\n%s",
                     self.endpoint, self._pretty_print_query(full_query),
                     "=" * 40)
        async with self.post(self.endpoint, data={"query": full_query},
                             headers=headers) as resp:
            try:
                resp.raise_for_status()
            except aiohttp.client_exceptions.ClientResponseError as exc:
                explanation = await resp.text()
                logger.debug("Server responded:\n%s\n%s",
                             explanation, "=" * 40)
                raise SPARQLRequestFailed(
                    exc.request_info, exc.history,
                    code=exc.code, message=exc.message,
                    explanation=explanation)
            return await resp.json()

    async def update(self, query: str, *args, **keywords) -> dict:
        headers = {
            "Accept": "application/json",
        }
        full_query = self._prepare_query(query, *args, **keywords)
        logger.debug("Sending SPARQL query to %s:\n%s\n%s",
                     self.endpoint, self._pretty_print_query(full_query),
                     "=" * 40)
        async with self.post(self.update_endpoint, data={"update": full_query},
                             headers=headers) as resp:
            try:
                resp.raise_for_status()
            except aiohttp.client_exceptions.ClientResponseError as exc:
                explanation = await resp.text()
                logger.debug("Server responded:\n%s\n%s",
                             explanation, "=" * 40)
                raise SPARQLRequestFailed(
                    exc.request_info, exc.history,
                    code=exc.code, message=exc.message, headers=exc.headers,
                    explanation=explanation)
            return await resp.json()



from aiohttp.test_utils import AioHTTPTestCase, TestClient, unittest_run_loop
from aiosparql.client import SPARQLClient
from aiosparql.syntax import IRI
import asyncio


__all__ = ['unittest_run_loop', 'AioSPARQLTestCase', 'TestSPARQLClient']


class TestSPARQLClient(TestClient, SPARQLClient):
    def __init__(self, app, **kwargs):
        TestClient.__init__(self, app, loop=kwargs.get('loop'))
        SPARQLClient.__init__(self, **kwargs)


class AioSPARQLTestCase(AioHTTPTestCase):
    client_kwargs = {
        "endpoint": "/sparql",
        "graph": IRI("http://mu.semte.ch/test-application"),
    }

    @asyncio.coroutine
    def _get_client(self, app):
        return TestSPARQLClient(app, loop=self.loop, **self.client_kwargs)



#!/usr/bin/env python3

import os

from setuptools import setup, find_packages

ROOT_DIR = os.path.dirname(__file__)
SOURCE_DIR = os.path.join(ROOT_DIR)

version = None
exec(open('aiosparql/__init__.py').read())

with open('./requirements.txt') as reqs_txt:
    requirements = list(iter(reqs_txt))


with open('./test-requirements.txt') as test_reqs_txt:
    test_requirements = list(iter(test_reqs_txt))


setup(
    name="aiosparql",
    version=version,
    description="An asynchronous SPARQL library using aiohttp",
    url='https://github.com/tenforce/sparql-aiohttp',
    packages=find_packages(exclude=["tests.*", "tests"]),
    install_requires=requirements,
    tests_require=test_requirements,
    zip_safe=False,
    test_suite='tests',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
        'License :: OSI Approved :: Apache Software License',
    ],
    maintainer='Cecile Tonglet',
    maintainer_email='cecile.tonglet@gmail.com',
)
